#include "common.h"
#include <stdio.h>
#include "processor.h"
#include "SystemInfo.h"
#include "Label.h"
#include "NetStats.h"
#include "wa3remote.h"

// FIXME: Must be a better way to do this
#define DEFAULT_FILE_DATE_FORMAT "dd mmm yyyy hh:nn:ss"

const int unitsDefault = 0;
const int unitsBytes = 1;
const int unitsKB = 2;
const int unitsMB = 3;
const int unitsGB = 4;
const int unitsPercent = 5;

//    enum Bus
#define BusType     char
#define ISA         0
#define SMBus       1
#define VIA686Bus   2
#define DirectIO    3

//    enum SMB
#define SMBType         char
#define smtSMBIntel     0
#define smtSMBAMD       1
#define smtSMBALi       2
#define smtSMBNForce    3
#define smtSMBSIS       4


// enum Sensor Types
#define SensorType      char
#define stUnknown       0
#define stTemperature   1
#define stVoltage       2
#define stFan           3
#define stMhz           4
#define stPercentage    5  


typedef struct {
    SensorType  iType;          // type of sensor
    int         Count;          // number of sensor for that type
} SharedIndex;

typedef struct {
    SensorType ssType;          // type of sensor
    unsigned char ssName[12];   // name of sensor
    char sspadding1[3];         // padding of 3 byte
    double ssCurrent;           // current value
    double ssLow;               // lowest readout
    double ssHigh;              // highest readout
    long ssCount;               // total number of readout
    char sspadding2[4];         // padding of 4 byte
    long double ssTotal;        // total amout of all readouts
    char sspadding3[6];         // padding of 6 byte
    double ssAlarm1;            // temp & fan: high alarm; voltage: % off;
    double ssAlarm2;            // temp: low alarm
} SharedSensor;

typedef struct {
    short siSMB_Base;            // SMBus base address
    BusType siSMB_Type;         // SMBus/Isa bus used to access chip
    SMBType siSMB_Code;         // SMBus sub type, Intel, AMD or ALi
    char siSMB_Addr;            // Address of sensor chip on SMBus
    unsigned char siSMB_Name[41];        // Nice name for SMBus
    short siISA_Base;            // ISA base address of sensor chip on ISA
    int siChipType;             // Chip nr, connects with Chipinfo.ini
    char siVoltageSubType;      // Subvoltage option selected
} SharedInfo;

typedef struct {
    double sdVersion;           // version number (example: 51090)
    SharedIndex sdIndex[10];     // Sensor index
    SharedSensor sdSensor[100];  // sensor info
    SharedInfo sdInfo;          // misc. info
    unsigned char sdStart[41];           // start time                              
    unsigned char sdCurrent[41];         // current time
    unsigned char sdPath[256];           // MBM path
} mbmSharedData;


SystemInfo::SystemInfo()
{
	WSADATA wd;
	WSAStartup(0x0101, &wd);
	usingCPU = false;

	maskStack.push_back(true);

	am = getLocaleInfo(LOCALE_S1159);
	pm = getLocaleInfo(LOCALE_S2359);

	shortWeekdays[0] = getLocaleInfo(LOCALE_SABBREVDAYNAME7);
	shortWeekdays[1] = getLocaleInfo(LOCALE_SABBREVDAYNAME1);
	shortWeekdays[2] = getLocaleInfo(LOCALE_SABBREVDAYNAME2);
	shortWeekdays[3] = getLocaleInfo(LOCALE_SABBREVDAYNAME3);
	shortWeekdays[4] = getLocaleInfo(LOCALE_SABBREVDAYNAME4);
	shortWeekdays[5] = getLocaleInfo(LOCALE_SABBREVDAYNAME5);
	shortWeekdays[6] = getLocaleInfo(LOCALE_SABBREVDAYNAME6);

	weekdays[0] = getLocaleInfo(LOCALE_SDAYNAME7);
	weekdays[1] = getLocaleInfo(LOCALE_SDAYNAME1);
	weekdays[2] = getLocaleInfo(LOCALE_SDAYNAME2);
	weekdays[3] = getLocaleInfo(LOCALE_SDAYNAME3);
	weekdays[4] = getLocaleInfo(LOCALE_SDAYNAME4);
	weekdays[5] = getLocaleInfo(LOCALE_SDAYNAME5);
	weekdays[6] = getLocaleInfo(LOCALE_SDAYNAME6);

	shortMonths[0] = getLocaleInfo(LOCALE_SABBREVMONTHNAME1);
	shortMonths[1] = getLocaleInfo(LOCALE_SABBREVMONTHNAME2);
	shortMonths[2] = getLocaleInfo(LOCALE_SABBREVMONTHNAME3);
	shortMonths[3] = getLocaleInfo(LOCALE_SABBREVMONTHNAME4);
	shortMonths[4] = getLocaleInfo(LOCALE_SABBREVMONTHNAME5);
	shortMonths[5] = getLocaleInfo(LOCALE_SABBREVMONTHNAME6);
	shortMonths[6] = getLocaleInfo(LOCALE_SABBREVMONTHNAME7);
	shortMonths[7] = getLocaleInfo(LOCALE_SABBREVMONTHNAME8);
	shortMonths[8] = getLocaleInfo(LOCALE_SABBREVMONTHNAME9);
	shortMonths[9] = getLocaleInfo(LOCALE_SABBREVMONTHNAME10);
	shortMonths[10] = getLocaleInfo(LOCALE_SABBREVMONTHNAME11);
	shortMonths[11] = getLocaleInfo(LOCALE_SABBREVMONTHNAME12);

	months[0] = getLocaleInfo(LOCALE_SMONTHNAME1);
	months[1] = getLocaleInfo(LOCALE_SMONTHNAME2);
	months[2] = getLocaleInfo(LOCALE_SMONTHNAME3);
	months[3] = getLocaleInfo(LOCALE_SMONTHNAME4);
	months[4] = getLocaleInfo(LOCALE_SMONTHNAME5);
	months[5] = getLocaleInfo(LOCALE_SMONTHNAME6);
	months[6] = getLocaleInfo(LOCALE_SMONTHNAME7);
	months[7] = getLocaleInfo(LOCALE_SMONTHNAME8);
	months[8] = getLocaleInfo(LOCALE_SMONTHNAME9);
	months[9] = getLocaleInfo(LOCALE_SMONTHNAME10);
	months[10] = getLocaleInfo(LOCALE_SMONTHNAME11);
	months[11] = getLocaleInfo(LOCALE_SMONTHNAME12);

	netStats = new NetStats;
	netStats->Start();
}

SystemInfo::~SystemInfo()
{
	netStats->Stop();
	delete netStats;

	if(usingCPU)
		endProcessorStatistics();

	WSACleanup();

	delete[] am;
	delete[] pm;

	int i;

	for(i = 0; i < 7; i++)
	{
		delete[] shortWeekdays[i];
		delete[] weekdays[i];
	}

	for(i = 0; i < 12; i++)
	{
		delete[] shortMonths[i];
		delete[] months[i];
	}
}

string SystemInfo::processLabelText(const string &text, Label *label, bool *dynamic)
{
	string output0, output1;
	
	int length = text.length();
	int i = 0;
	
	if(dynamic)
		*dynamic = false;
	
	while(i < length)
	{
		if(text[i] == '[')
		{
			i++;
			
			if(i < length && text[i] == '[')
			{
				if(maskStack.back()) output0.append(1, '[');
				i++;
				continue;
			}
			
			string temp = process(text, i, length, label, dynamic);
			if(maskStack.back()) output0.append(temp);

			if(i >= length || text[i++] != ']')
			{
				output0 = "[syntax error]";
				break;
			}
			continue;
		}
		else if(text[i] == ']')
		{
			i++;
			
			if(i < length && text[i] == ']')
			{
				if(maskStack.back()) output0.append(1, ']');
				i++;
				continue;
			}
		}
		else if(text[i] == '\\')
		{
			char ch;
			i++;

			if(i < length && text[i] == 'n')
			{
				ch = '\n';
			}
			else
			{
				ch = text[i];
			}

			if(maskStack.back()) output0.append(1, ch);
			i++;
			continue;
		}
		
		if(maskStack.back()) output0.append(1, text[i]);
		i++;
	}
	
	length = output0.length();
	i = 0;
	
	while(i < length)
	{
		if(output0[i] == '\\')
		{
			char ch;
			i++;

			if(i < length && output0[i] == 'n')
			{
				ch = '\n';
			}
            else
			{
				ch = output0[i];
			}

			if(maskStack.back()) output1.append(1, ch);
			i++;
			continue;
		}
		
		if(maskStack.back()) output1.append(1, output0[i]);
		i++;
	}

	return output1;
}

string SystemInfo::evaluateFunction(const string &functionName, const vector<string> &arguments, Label *label, bool *dynamic = 0)
{
	string name = lowerCase(functionName);
	
	if(name == "if" && arguments.size() >= 1)
	{
		doIf(arguments[0]);
		return "";
	}
	else if(name == "elseif" && arguments.size() >= 1)
	{
		doElseIf(arguments[0]);
		return "";
	}
	else if(name == "else")
	{
		doElse();
		return "";
	}
	else if(name == "endif")
	{
		doEndIf();
		return "";
	}

	if(name == "activetask")
		return getActiveTask(dynamic);
	else if(name == "battery")
		return getBattery(dynamic);
	else if(name == "computername")
		return getComputerName(dynamic);
	else if(name == "kblayout")
		return getKeyboardLayoutName(dynamic);
	else if(name == "skblayout")
		return getKeyboardLayoutShortName(dynamic);
	else if(name == "connected")
		return getConnected(dynamic);
	else if(name == "cpu")
		return getCPU(dynamic);
	else if(name == "date")
		return getDate(arguments, dynamic);
	else if(name == "datecreated")
		return getDateCreated(arguments, dynamic);
	else if(name == "datelastmodified")
		return getDateLastModified(arguments, dynamic);
	else if(name == "diskavailable" && arguments.size() >= 1)
		return getDiskAvailable(arguments, dynamic);
	else if(name == "diskinuse" && arguments.size() >= 1)
		return getDiskInUse(arguments, dynamic);
	else if(name == "disktotal" && arguments.size() >= 1)
		return getDiskTotal(arguments, dynamic);
	else if(name == "fileexists" && arguments.size() >= 1)
		return getFileExists(arguments[0], dynamic);
	else if(name == "firstline" && arguments.size() >= 1)
		return getFirstLine(arguments[0], dynamic);
	else if(name == "hostname")
		return getHostName(dynamic);
	else if(name == "ip")
		return getIP(arguments, dynamic);
	else if(name == "itime")
		return getInternetTime(dynamic);
	else if(name == "lastline" && arguments.size() >= 1)
		return getLastLine(arguments[0], dynamic);
	else if(name == "linecount" && arguments.size() >= 1)
		return getLineCount(arguments[0], dynamic);
	else if(name == "line")
		return getLine(arguments, dynamic);
	else if(name == "mbmtemperature")
		return getMBMData(stTemperature, arguments, dynamic, label->bUseFahrenheit);
	else if(name == "mbmcpuusage")
		return getMBMData(stPercentage, arguments, dynamic, label->bUseFahrenheit);
	else if(name == "mbmfanspeed")
		return getMBMData(stFan, arguments, dynamic, label->bUseFahrenheit);
	else if(name == "mbmvoltage")
		return getMBMData(stVoltage, arguments, dynamic, label->bUseFahrenheit);
	else if(name == "mbmloaded")
		return getMBMLoaded(dynamic);
	else if(name == "memavailable")
		return getMemAvailable(arguments, dynamic);
	else if(name == "meminuse")
		return getMemInUse(arguments, dynamic);
	else if(name == "memtotal")
		return getMemTotal(arguments, dynamic);
	else if(name == "netin")
		return getNetIn(arguments, dynamic);
	else if(name == "netout")
		return getNetOut(arguments, dynamic);
	else if(name == "os")
		return getOS(dynamic);
	else if(name == "powersource")
		return getPowerSource(arguments, dynamic);
	else if(name == "randomline" && arguments.size() >= 1)
		return getRandomLine(arguments[0], dynamic);
	else if(name == "size")
		return getSize(arguments, dynamic);
	else if(name == "swapavailable")
		return getSwapAvailable(arguments, dynamic);
	else if(name == "swapinuse")
		return getSwapInUse(arguments, dynamic);
	else if(name == "swaptotal")
		return getSwapTotal(arguments, dynamic);
	else if(name == "time")
		return getTime(arguments, dynamic);
	else if(name == "uptime")
		return getUptime(arguments, dynamic);
	else if(name == "username")
		return getUserName(dynamic);
	else if(name == "winampsong")
		return getWinampSong(dynamic);
	else if(name == "winamptime")
		return getWinampTime(dynamic);
	else if(name == "winampremaintime")
		return getWinampRemainTime(dynamic);
	else if(name == "winamptotaltime")
		return getWinampTotalTime(dynamic);
	else if(name == "winampstatus")
		return getWinampStatus(dynamic);
	else if(name == "winampplaying")
		return getWinampPlaying(dynamic);
	else if(name == "windowtitle" && arguments.size() >= 1)
		return getWindowTitle(arguments[0], dynamic);
	else if (name == "clipboardtext")
		return getClipboardText(dynamic);
	else if(name == "hideifempty" && arguments.size() >= 1)
		return hideIfEmpty(arguments[0], label, dynamic);
	else if(name == "empty" && arguments.size() == 1)
		return empty(arguments[0], dynamic);
	else if(name == "notempty" && arguments.size() == 1)
		return notEmpty(arguments[0], dynamic);
	else if(name == "after" && arguments.size() == 2)
		return after(arguments[0], arguments[1]);
	else if(name == "afterlast" && arguments.size() == 2)
		return afterLast(arguments[0], arguments[1]);
	else if(name == "before" && arguments.size() == 2)
		return before(arguments[0], arguments[1]);
	else if(name == "beforelast" && arguments.size() == 2)
		return beforeLast(arguments[0], arguments[1]);
	else if(name == "between" && arguments.size() == 3)
		return between(arguments[0], arguments[1], arguments[2]);
	else if(name == "capitalize" && arguments.size() == 1)
		return capitalize(arguments[0]);
	else if(name == "lowercase" && arguments.size() == 1)
		return lowerCase(arguments[0]);
	else if(name == "trim" && arguments.size() == 1)
		return trim(arguments[0]);
	else if(name == "uppercase" && arguments.size() == 1)
		return upperCase(arguments[0]);
	else if(name == "verticaldown" && arguments.size() == 1)
		return verticaldown(arguments[0]);
	else if(name == "verticalup" && arguments.size() == 1)
		return verticalup(arguments[0]);
	else if(name == "bytes") // units hack
		return "1";
	else if(name == "kb")
		return "2";
	else if(name == "mb")
		return "3";
	else if(name == "gb")
		return "4";
	else if(name == "%")
		return "5";
	else if(name == "semicolon") // hack to avoid escaping ;'s
		return ";";

	*dynamic = false;
	return "[?]";
}

string SystemInfo::process(const string &text, int &i, int length, Label *label, bool *dynamic = 0)
{
	// skip whitespace
	while(i < length && isspace(text[i]))
		i++;
	
	// is it a string literal?
	if(i < length && (text[i] == '\"' || text[i] == '\''))
	{
		string literalValue;
		char quoteChar = text[i++];
		
		while(i < length && text[i] != quoteChar)
			literalValue.append(1, text[i++]);
		
		if(i < length) i++;
		return literalValue;
	}
	else if(i < length && isdigit(text[i])) // is it a numeric constant?
	{
		string number;

		while(i < length && isdigit(text[i]))
			number.append(1, text[i++]);

		return number;
	}
	else // otherwise its a name
	{
		string name;
		vector<string> arguments;
		
		// retrieve name
		while(i < length && (isalnum(text[i]) || text[i] == '-' || text[i] == '%'))
			name.append(1, text[i++]);
		
		// skip whitespace
		while(i < length && isspace(text[i]))
			i++;
		
		// has argument list?
		if(i < length && text[i] == '(')
		{
			// process each argument
			do
			{
				i++;

				// call recursively to handle
				arguments.push_back(process(text, i, length, label, dynamic));
				
				// skip whitespace
				while(i < length && isspace(text[i]))
					i++;
			}
			while(i < length && text[i] == ',');

			// must be a closing parenthesis
			if(i >= length || text[i++] != ')')
				return "[error]";
		}
		
		// get it
		return evaluateFunction(name, arguments, label, dynamic);
	}
}

void SystemInfo::doIf(const string& test)
{
	condStack.push_back(atoi(test.c_str()) != 0);
	depthStack.push_back(1);
	maskStack.push_back(maskStack.back() && condStack.back());
}

void SystemInfo::doElseIf(const string& test)
{
	if(maskStack.empty())
		return;

	maskStack.pop_back();
	maskStack.push_back(maskStack.back() && !condStack.back());
	condStack.push_back(atoi(test.c_str()) != 0);
	depthStack.back()++;
	maskStack.push_back(maskStack.back() && condStack.back());
}

void SystemInfo::doElse()
{
	if(maskStack.empty())
		return;

	maskStack.pop_back();
	maskStack.push_back(maskStack.back() && !condStack.back());
}

void SystemInfo::doEndIf()
{
	if(depthStack.empty())
		return;

	int depth = depthStack.back();
	depthStack.pop_back();

	while(depth--)
	{
		condStack.pop_back();
		maskStack.pop_back();
	}
}

string SystemInfo::getActiveTask(bool *dynamic)
{
	char buffer[256] = { 0 };
	HWND activeTask = GetForegroundWindow();

	if(IsAppWindow(activeTask))
	{
		GetWindowText(activeTask, buffer, 256);
		if(dynamic) *dynamic = true;
			return string(buffer);
	}
	else
	{
		if(dynamic) *dynamic = true;
			return "Unknown Task";
	}
}

string SystemInfo::getComputerName(bool *dynamic)
{
	char buffer[32];
	unsigned long bufferLen = 32;

	GetComputerName(buffer, &bufferLen);
	return string(buffer);
}

string SystemInfo::getKeyboardLayoutName(bool *dynamic)
{
	char buffer[128];

	DWORD idthd=GetWindowThreadProcessId(GetForegroundWindow(),NULL);
	GetLocaleInfo(MAKELCID(LOWORD(GetKeyboardLayout(idthd)), SORT_DEFAULT),LOCALE_SLANGUAGE,buffer,128);
	return string(buffer);
}

string SystemInfo::getKeyboardLayoutShortName(bool *dynamic)
{
	char buffer[128];

	DWORD idthd=GetWindowThreadProcessId(GetForegroundWindow(),NULL);
	GetLocaleInfo(MAKELCID(LOWORD(GetKeyboardLayout(idthd)), SORT_DEFAULT),LOCALE_SISO639LANGNAME,buffer,128);
	return string(buffer);

}

string SystemInfo::getConnected(bool *dynamic)
{
	if(dynamic) *dynamic = true;
	return netStats->IsConnAvailable() ? "1" : "0";
}

string SystemInfo::getCPU(bool *dynamic)
{
	if(!usingCPU)
	{
		startProcessorStatistics();
		usingCPU = true;
	}

	char buffer[16];
	StringCchPrintf(buffer, 16, "%d%%", getProcessorUsage());

	if(dynamic) *dynamic = true;
	return string(buffer);
}

string SystemInfo::getDate(const vector<string> &arguments, bool *dynamic)
{
	if(dynamic) *dynamic = true;

	SYSTEMTIME st;
	getLocalizedTime(&st, arguments.size() >= 2 ? arguments[1] : "");

	if(arguments.size() >= 1)
	{
		// SYSTEMTIME st;
		// GetLocalTime(&st);
		return formatDateTime(arguments[0], st, 0);
	}
	else
	{
		char buffer[64];
		GetDateFormat(LOCALE_USER_DEFAULT, DATE_LONGDATE, &st, 0, buffer, 64);
		return string(buffer);
	}
}

string SystemInfo::getInternetTime(bool *dynamic)
{
	if(dynamic) *dynamic = true;

	SYSTEMTIME st;
	GetSystemTime(&st);

	int seconds = ((((st.wHour + 1) % 24) * 60) + st.wMinute) * 60 + st.wSecond;
	int beats = (1000 * seconds) / 86400;

	char output[8];
	StringCchPrintf(output, 8, "@%03d", beats);
	return string(output);
}

string SystemInfo::getDiskAvailable(const vector<string> &arguments, bool *dynamic)
{
	largeInt freeBytes;
	largeInt totalBytes;
	int units = unitsDefault;

	if(arguments.size() >= 2)
		units = atoi(arguments[1].c_str());

	getDiskFreeSpace(arguments[0], freeBytes, totalBytes);
	
	if(dynamic) *dynamic = true;
	return formatByteSize(freeBytes, totalBytes, units);
}

string SystemInfo::getDiskInUse(const vector<string> &arguments, bool *dynamic)
{
	largeInt freeBytes;
	largeInt totalBytes;
	int units = unitsDefault;

	if(arguments.size() >= 2)
		units = atoi(arguments[1].c_str());

	getDiskFreeSpace(arguments[0], freeBytes, totalBytes);
	
	if(dynamic) *dynamic = true;
	return formatByteSize(totalBytes - freeBytes, totalBytes, units);
}

string SystemInfo::getDiskTotal(const vector<string> &arguments, bool *dynamic)
{
	largeInt freeBytes;
	largeInt totalBytes;
	int units = unitsDefault;

	if(arguments.size() >= 2)
		units = atoi(arguments[1].c_str());

	getDiskFreeSpace(arguments[0], freeBytes, totalBytes);
	
	if(dynamic) *dynamic = true;
	return formatByteSize(totalBytes, totalBytes, units);
}

string SystemInfo::getHostName(bool *dynamic)
{
	char hostName[256];
	gethostname(hostName, 256);
	hostent *hostInfo = gethostbyname(hostName);

	if(hostInfo)
		return string(hostInfo->h_name);
	
	return "";
}

string SystemInfo::getIP(const vector<string> &arguments, bool *dynamic)
{
	char hostName[256];
	gethostname(hostName, 256);
	hostent *hostInfo = gethostbyname(hostName);

	if(!hostInfo)
		return "";

	int numIPs = 0;

	while(hostInfo->h_addr_list[numIPs])
		numIPs++;

	if(numIPs == 0)
		return "";

	int n = 0;
	int value = 0;

	if(arguments.size() >= 1)
		n = atoi(arguments[0].c_str()) - 1;

	if(n < 0) n = 0;
	if(n >= numIPs) n = numIPs - 1;

	return string(inet_ntoa(*((in_addr *) hostInfo->h_addr_list[n])));
}

string SystemInfo::getMemAvailable(const vector<string> &arguments, bool *dynamic)
{
	int units = unitsDefault;

	if(arguments.size() >= 1)
		units = atoi(arguments[0].c_str());

	MEMORYSTATUS ms;
	ms.dwLength = sizeof(MEMORYSTATUS);
	GlobalMemoryStatus(&ms);

	if(dynamic) *dynamic = true;
	return formatByteSize(ms.dwAvailPhys, ms.dwTotalPhys, units);
}

string SystemInfo::getMemInUse(const vector<string> &arguments, bool *dynamic)
{
	int units = unitsDefault;

	if(arguments.size() >= 1)
		units = atoi(arguments[0].c_str());

	MEMORYSTATUS ms;
	ms.dwLength = sizeof(MEMORYSTATUS);
	GlobalMemoryStatus(&ms);

	if(dynamic) *dynamic = true;
	return formatByteSize(ms.dwTotalPhys - ms.dwAvailPhys, ms.dwTotalPhys, units);
}

string SystemInfo::getMemTotal(const vector<string> &arguments, bool *dynamic)
{
	int units = unitsDefault;

	if(arguments.size() >= 1)
		units = atoi(arguments[0].c_str());

	MEMORYSTATUS ms;
	ms.dwLength = sizeof(MEMORYSTATUS);
	GlobalMemoryStatus(&ms);

	if(dynamic) *dynamic = true;
	return formatByteSize(ms.dwTotalPhys, ms.dwTotalPhys, units);
}

string SystemInfo::getOS(bool *dynamic)
{
	if(IsOS(OS_ME))
		return "Windows ME";
	if(IsOS(OS_98))
		return "Windows 98";
	if(IsOS(OS_95))
		return "Windows 95";

	if(IsOS(OS_XP))
		return "Windows XP";
	if(IsOS(OS_2K))
		return "Windows 2000";
	if(IsOS(OS_NT4))
		return "Windows NT4";

	return "Unknown OS";
}

string SystemInfo::getDateCreated(const vector<string> &arguments, bool *dynamic)
{
	if(arguments.size() < 1)
		return "[error]";

	WIN32_FIND_DATA fd;
	HANDLE search = FindFirstFile(arguments[0].c_str(), &fd);
	FILETIME ftLocal;
	SYSTEMTIME st;

	if(search)
	{
		FindClose(search);
		FileTimeToLocalFileTime(&fd.ftCreationTime, &ftLocal);
		FileTimeToSystemTime(&ftLocal, &st);
		return formatDateTime((arguments.size() > 1) ? arguments[1] : DEFAULT_FILE_DATE_FORMAT, st, 0);
	}

	return "[Could not open file!]";
}

string SystemInfo::getDateLastModified(const vector<string> &arguments, bool *dynamic)
{
	if(arguments.size() < 1)
		return "[error]";

	WIN32_FIND_DATA fd;
	HANDLE search = FindFirstFile(arguments[0].c_str(), &fd);
	FILETIME ftLocal;
	SYSTEMTIME st;

	if(search)
	{
		FindClose(search);
		FileTimeToLocalFileTime(&fd.ftLastWriteTime, &ftLocal);
		FileTimeToSystemTime(&ftLocal, &st);
		return formatDateTime((arguments.size() > 1) ? arguments[1] : DEFAULT_FILE_DATE_FORMAT, st, 0);
	}

	return "[Could not open file!]";
}

string SystemInfo::getFileExists(const string &file, bool *dynamic)
{
	WIN32_FIND_DATA fd;
	HANDLE search = FindFirstFile(file.c_str(), &fd);

	if(search)
	{
		FindClose(search);
		return "1";
	}

	return "0";
}

string SystemInfo::getFirstLine(const string &file, bool *dynamic)
{
	FILE *fp = fopen(file.c_str(), "r");
	char lineBuffer[1024];

	if(fp)
	{
		if(fgets(lineBuffer, 1024, fp))
		{
			int length = strlen(lineBuffer);
			if(lineBuffer[length - 1] == '\n') lineBuffer[length - 1] = 0;
			fclose(fp);
			return string(lineBuffer);
		}

		fclose(fp);
		return "";
	}

	return "[Could not open file!]";
}

string SystemInfo::getLastLine(const string &file, bool *dynamic)
{
	FILE *fp = fopen(file.c_str(), "r");
	char lineBuffer[1024];

	if(fp)
	{
		strcpy(lineBuffer, "");

		while(fgets(lineBuffer, 1024, fp))
		{
			// Do nothing
		}

		int length = strlen(lineBuffer);
		if(lineBuffer[length - 1] == '\n') lineBuffer[length - 1] = 0;
		fclose(fp);
		return string(lineBuffer);
	}

	return "[Could not open file!]";
}

string SystemInfo::getLine(const vector<string> &arguments, bool *dynamic)
{
	if(arguments.size() < 2)
		return "[error]";

	string file = arguments[0];
	int lineNumber = atoi(arguments[1].c_str());
	FILE *fp = fopen(file.c_str(), "r");
	int lineCount = 0;
	char lineBuffer[1024];

	if(fp)
	{
		while(fgets(lineBuffer, 1024, fp))
		{
			if(lineCount == lineNumber - 1)
			{
				int length = strlen(lineBuffer);
				if(lineBuffer[length - 1] == '\n') lineBuffer[length - 1] = 0;
				fclose(fp);
				return string(lineBuffer);
			}

			lineCount++;
		}

		fclose(fp);
		return "";
	}

	return "[Could not open file!]";
}

string SystemInfo::getLineCount(const string &file, bool *dynamic)
{
	FILE *fp = fopen(file.c_str(), "r");
	int lineCount = 0;
	char lineBuffer[1024];

	if(fp)
	{
		while(fgets(lineBuffer, 1024, fp))
		{
			lineCount++;
		}

		fclose(fp);
		sprintf(lineBuffer, "%d", lineCount);
		return string(lineBuffer);
	}

	return "[Could not open file!]";
}

string SystemInfo::getRandomLine(const string &file, bool *dynamic)
{
	FILE *fp = fopen(file.c_str(), "r");
	int lineCount = 0;
	char lineBuffer[1024];

	if(fp)
	{
		while(fgets(lineBuffer, 1024, fp))
		{
			if(lineBuffer[0] != '\n' && lineBuffer[0] != 0)
				lineCount++;
		}

		srand(GetTickCount());
		int line = rand() % lineCount;

		lineCount = 0;
		fseek(fp, 0, SEEK_SET);

		while(fgets(lineBuffer, 1024, fp))
		{
			if(lineBuffer[0] != '\n' && lineBuffer[0] != 0)
			{
				if(lineCount == line)
				{
					int length = strlen(lineBuffer);
					if(lineBuffer[length - 1] == '\n') lineBuffer[length - 1] = 0;
					fclose(fp);
					return string(lineBuffer);
				}

				lineCount++;
			}
		}

		fclose(fp);
		return "[error]";
	}

	return "[Could not open file!]";
}

string SystemInfo::getSize(const vector<string> &arguments, bool *dynamic)
{
	if(arguments.size() < 1)
		return "[error]";

	int units = unitsDefault;

	if(arguments.size() >= 2)
		units = atoi(arguments[1].c_str());

	WIN32_FIND_DATA fd;
	HANDLE search = FindFirstFile(arguments[0].c_str(), &fd);

	if(search)
	{
		FindClose(search);
		largeInt size = (((largeInt) fd.nFileSizeHigh) << 32) | ((largeInt) fd.nFileSizeLow);
		return formatByteSize(size, 0, units);
	}

	return "[Could not open file!]";
}

string SystemInfo::getSwapAvailable(const vector<string> &arguments, bool *dynamic)
{
	int units = unitsDefault;

	if(arguments.size() >= 1)
		units = atoi(arguments[0].c_str());

	MEMORYSTATUS ms;
	ms.dwLength = sizeof(MEMORYSTATUS);
	GlobalMemoryStatus(&ms);

	if(dynamic) *dynamic = true;
	return formatByteSize(ms.dwAvailPageFile, ms.dwTotalPageFile, units);
}

string SystemInfo::getSwapInUse(const vector<string> &arguments, bool *dynamic)
{
	int units = unitsDefault;

	if(arguments.size() >= 1)
		units = atoi(arguments[0].c_str());

	MEMORYSTATUS ms;
	ms.dwLength = sizeof(MEMORYSTATUS);
	GlobalMemoryStatus(&ms);

	if(dynamic) *dynamic = true;
	return formatByteSize(ms.dwTotalPageFile - ms.dwAvailPageFile, ms.dwTotalPageFile, units);
}

string SystemInfo::getSwapTotal(const vector<string> &arguments, bool *dynamic)
{
	int units = unitsDefault;

	if(arguments.size() >= 1)
		units = atoi(arguments[0].c_str());

	MEMORYSTATUS ms;
	ms.dwLength = sizeof(MEMORYSTATUS);
	GlobalMemoryStatus(&ms);

	if(dynamic) *dynamic = true;
	return formatByteSize(ms.dwTotalPageFile, ms.dwTotalPageFile, units);
}

string SystemInfo::getTime(const vector<string> &arguments, bool *dynamic)
{
	if(dynamic) *dynamic = true;

	SYSTEMTIME st;
	getLocalizedTime(&st, arguments.size() >= 2 ? arguments[1] : "");

	if(arguments.size() >= 1)
	{
		// SYSTEMTIME st;
		// GetLocalTime(&st);
		return formatDateTime(arguments[0], st, 0);
	}
	else
	{
		char buffer[64];
		GetTimeFormat(LOCALE_USER_DEFAULT, TIME_NOSECONDS, &st, 0, buffer, 64);
		return string(buffer);
	}
}

string SystemInfo::getUptime(const vector<string> &arguments, bool *dynamic)
{
	if(dynamic) *dynamic = true;

	unsigned int ms = GetTickCount();
	unsigned int seconds = ms / 1000;
	ms %= 1000;
	unsigned int minutes = seconds / 60;
	seconds %= 60;
	unsigned int hours = minutes / 60;
	minutes %= 60;
	unsigned int days = hours / 24;
	hours %= 24;

	SYSTEMTIME st;

	st.wMonth = -1;
	st.wDay = days;
	st.wDayOfWeek = -1;
	st.wYear = 0;
	st.wHour = hours;
	st.wMinute = minutes;
	st.wSecond = seconds;
	st.wMilliseconds = ms;

	string format = (arguments.size() >= 1) ? arguments[0] : "d days hh:nn:ss";
	return formatDateTime(format, st, 1);
}

string SystemInfo::getUserName(bool *dynamic)
{
	char buffer[32];
	unsigned long bufferLen = 32;

	GetUserName(buffer, &bufferLen);
	return string(buffer);
}

string SystemInfo::getWinampSong(bool *dynamic)
{
	if(dynamic) *dynamic = true;

	HWND hwndWA2 = FindWindow("Winamp v1.x", NULL);
	HWND hwndWA3 = FindWindow(WA3REMOTE_WNDCLASS, NULL);
	HWND hwndQCD = FindWindow("PlayerCanvas", NULL);
	HWND hwndFoo2K = FindWindow("FOOBAR2000_CLASS", NULL);

	if (hwndFoo2K != NULL)
	{
		char buffer[256];
		GetWindowText(hwndFoo2K, buffer, 256);
		
		return trim(before(string(buffer), "<>"));
	}
	else if(hwndQCD != NULL)
	{
		// QuintEssential Player
		char buffer[256];
		GetWindowText(hwndQCD, buffer, 256);
		
		return trim(afterLast(string(buffer), "]"));
	}
	else if(hwndWA2 != NULL)
	{
		// WA2
		char buffer[256];
		GetWindowText(hwndWA2, buffer, 256);

		if(strstr(buffer, " - Winamp") == 0)
			return "";

		return trim(between(string(buffer), ".", "-"));
	}
	else if(hwndWA3 != NULL)
	{
		// WA3 (wa3remote.wac)
		HANDLE memHandle = OpenFileMapping(FILE_MAP_READ, FALSE, WA3REMOTE_SHAREDMEMNAME);
		PWA3REMOTE_SHAREDMEM memPtr;
		string result;

		if(memHandle)
		{
			memPtr = (PWA3REMOTE_SHAREDMEM) MapViewOfFile(memHandle, FILE_MAP_READ, 0, 0, 0);

			if(memPtr)
			{
				result = memPtr->szName;
				UnmapViewOfFile(memPtr);
			}

			CloseHandle(memHandle);
		}

		return result;
	}

	return "Player Not Running";
}

string SystemInfo::getWinampTime(bool *dynamic)
{
	if(dynamic) *dynamic = true;

	HWND hwndWA2 = FindWindow("Winamp v1.x", NULL);
	HWND hwndWA3 = FindWindow(WA3REMOTE_WNDCLASS, NULL);
	HWND hwndFoo2K = FindWindow("FOOBAR2000_CLASS", NULL);
	int milliseconds = 0;

	if (hwndFoo2K != NULL)
	{
		char foobuffer[256];
		GetWindowText(hwndFoo2K, foobuffer, 256);

		string result = after(string(foobuffer), "Time:");
		
		return trim(before(string(result), "/"));
	}
	else if(hwndWA2)
	{
		// WA2
		milliseconds = SendMessage(hwndWA2, WM_USER, 0, 105);
	}
	else if(hwndWA3)
	{
		// WA3 (wa3remote.wac)
		milliseconds = SendMessage(hwndWA3, WA3REMOTE_WM_GETPOSITION, 0, 0);
	}

	int seconds = milliseconds / 1000;
	int minutes = seconds / 60;
	seconds = seconds % 60;

	char buffer[80];
	sprintf(buffer, "%d:%02d", minutes, seconds);
	if (milliseconds != 0)
		return string(buffer);
	else
		return "";
}

string SystemInfo::getWinampRemainTime(bool *dynamic)
{
	if(dynamic) *dynamic = true;

	HWND hwndWA2 = FindWindow("Winamp v1.x", NULL);
	HWND hwndWA3 = FindWindow(WA3REMOTE_WNDCLASS, NULL);
	HWND hwndFoo2K = FindWindow("FOOBAR2000_CLASS", NULL);
	int milliseconds = 0;

	if (hwndFoo2K != NULL)
	{
		char foobuffer[256];
		GetWindowText(hwndFoo2K, foobuffer, 256);

		string result = after(string(foobuffer), "Time:");
		result = after(string(result), "Time:");
		
		return trim(result);
	}
	else if(hwndWA2)
	{
		// WA2
		milliseconds = (SendMessage(hwndWA2, WM_USER, 1, 105) * 1000) - SendMessage(hwndWA2, WM_USER, 0, 105);
	}
	else if(hwndWA3)
	{
		// WA3 (wa3remote.wac)
		milliseconds = SendMessage(hwndWA3, WA3REMOTE_WM_GETLENGTH, 0, 0) - SendMessage(hwndWA3, WA3REMOTE_WM_GETPOSITION, 0, 0);
	}

	int seconds = milliseconds / 1000;
	int minutes = seconds / 60;
	seconds = seconds % 60;

	char buffer[80];
	sprintf(buffer, "%d:%02d", minutes, seconds);
	if (milliseconds != 0)
		return string(buffer);
	else
		return "";
}

string SystemInfo::getWinampTotalTime(bool *dynamic)
{
	if(dynamic) *dynamic = true;

	HWND hwndWA2 = FindWindow("Winamp v1.x", NULL);
	HWND hwndWA3 = FindWindow(WA3REMOTE_WNDCLASS, NULL);
	HWND hwndFoo2K = FindWindow("FOOBAR2000_CLASS", NULL);
	int milliseconds = 0;

	if (hwndFoo2K != NULL)
	{
		char foobuffer[256];
		GetWindowText(hwndFoo2K, foobuffer, 256);
		
		string result = afterLast(string(foobuffer), "/");

		return trim(before(result, " "));
	}
	else if(hwndWA2)
	{
		// WA2
		milliseconds = SendMessage(hwndWA2, WM_USER, 1, 105) * 1000;
	}
	else if(hwndWA3)
	{
		// WA3 (wa3remote.wac)
		milliseconds = SendMessage(hwndWA3, WA3REMOTE_WM_GETLENGTH, 0, 0);
	}

	int seconds = milliseconds / 1000;
	int minutes = seconds / 60;
	seconds = seconds % 60;

	char buffer[80];
	sprintf(buffer, "%d:%02d", minutes, seconds);
	if (milliseconds != 0)
		return string(buffer);
	else
		return "";
}

string SystemInfo::getWinampStatus(bool *dynamic)
{
	if(dynamic) *dynamic = true;

	HWND hwndWA2 = FindWindow("Winamp v1.x", NULL);
	HWND hwndWA3 = FindWindow(WA3REMOTE_WNDCLASS, NULL);

	if(hwndWA2)
	{
		// WA2
		int status = SendMessage(hwndWA2, WM_USER, 0, 104);

		switch(status)
		{
		case 1:
			return "Playing";

		case 3:
			return "Paused";

		default:
			return "Stopped";
		}
	}
	else if(hwndWA3)
	{
		// WA3 (wa3remote.wac)
		int status = SendMessage(hwndWA3, WA3REMOTE_WM_GETSTATUS, 0, 0);

		switch(status)
		{
		case 1:
			return "Playing";

		case -1:
			return "Paused";

		default:
			return "Stopped";
		}
	}

	return "Winamp Not Running";
}

string SystemInfo::getWinampPlaying(bool *dynamic)
{
	HWND hwndWA2 = FindWindow("Winamp v1.x", NULL);
	HWND hwndWA3 = FindWindow(WA3REMOTE_WNDCLASS, NULL);
	string retval = "0";

	if(hwndWA2)
	{
		// WA2
		if(SendMessage(hwndWA2, WM_USER, 0, 104) == 1)
			retval = "1";
	}
	else if(hwndWA3)
	{
		// WA3 (wa3remote.wac)
		if(SendMessage(hwndWA3, WA3REMOTE_WM_GETSTATUS, 0, 0) == 1)
			retval = "1";
	}

	if(dynamic) *dynamic = true;
	return retval;
}

string SystemInfo::getWindowTitle(const string &windowClass, bool *dynamic)
{
	if(dynamic) *dynamic = true;

	HWND hWnd = FindWindow(windowClass.c_str(), 0);
	if(!hWnd) return "";

	char buffer[256];
	GetWindowText(hWnd, buffer, 256);
	return string(buffer);
}

string SystemInfo::getPowerSource(const vector<string> &arguments, bool *dynamic)
{
	if(dynamic) *dynamic = 1;

	SYSTEM_POWER_STATUS sps;
	GetSystemPowerStatus(&sps);

	if(sps.ACLineStatus == 0)
	{
		if(arguments.size() >= 2)
			return arguments[1];
		else
			return "Battery";
	}
	else
	{
		if(arguments.size() >= 1)
			return arguments[0];
		else
			return "AC";
	}
}

string SystemInfo::getBattery(bool *dynamic)
{
	if(dynamic) *dynamic = 1;

	SYSTEM_POWER_STATUS sps;
	GetSystemPowerStatus(&sps);

	if(sps.BatteryLifePercent > 100)
		sps.BatteryLifePercent = 0;

	char output[8];
	StringCchPrintf(output, 8, "%d%%", sps.BatteryLifePercent);
	return string(output);
}

// -----------------------------------
//  MotherBoard Monitor 5 (MBM5) info
// -----------------------------------
/*
sensortype:		the type of sensor it should return (temp, voltage, fan, etc)
arguments:		# of sensor (you can have more than one temp sensor)
dynamic:		if this is a dynamic variable which can change over time
bUseFahrenheit:	if we should return the values in Fahrenheit
*/
string SystemInfo::getMBMData (const SensorType sensortype, const vector<string> &arguments, bool *dynamic, bool &bUseFahrenheit)
{
	if(dynamic) *dynamic = 1;	//this is always a dynamic variable

	int n = 0;
	double value = 0;
	mbmSharedData *pSharedData = NULL;
	char output[16];

	if(arguments.size() >= 1)
		n = atoi(arguments[0].c_str()) - 1;

	if(n < 0) n = 0;		//n has now the # of the sensor were looking for

	HANDLE hSharedData=OpenFileMapping(FILE_MAP_READ, FALSE, "$M$B$M$5$S$D$");
	if (hSharedData==0)
		return "0";
	
	if(hSharedData != INVALID_HANDLE_VALUE)
	{	pSharedData = (mbmSharedData *) MapViewOfFile(hSharedData,
			FILE_MAP_READ, 0, 0, 0);
	if(!pSharedData) {
			CloseHandle (hSharedData);
			UnmapViewOfFile(pSharedData);
			return "0";
		}
	}

	//this only works with newer versions of MBM
	if (pSharedData->sdVersion < 51090) {
		return "Please update your version of MBM";
	}

	int totalSensorCount = 0;
	for (int i=0;i<5;i++)
		totalSensorCount +=	pSharedData->sdIndex[i].Count;


	for (int j=0;j<totalSensorCount;j++)
	{	
		if (pSharedData->sdSensor[j].ssType == sensortype)	//is this the type of sensor were looking for?
		{
			if (n == 0)		//have we found our sensor?
			{
				switch (sensortype) 
				{
					case stTemperature:

						value = pSharedData->sdSensor[j].ssCurrent;
						if (bUseFahrenheit) {
							value = value * 9 / 5 + 32;
						} 

						StringCbPrintf (output, 15, "%.0f", value);
						break;
					case stVoltage:
					case stFan:
						value = pSharedData->sdSensor[j].ssCurrent;
						StringCbPrintf (output, 15, "%.2f", value);
						break;
				}
				UnmapViewOfFile(pSharedData);
				CloseHandle (hSharedData);
				return string(output);
			}
			else		//no we havent found it yet: same type, but wrong number
			{
				n--;
			}
		}
	}

	UnmapViewOfFile(pSharedData);
	CloseHandle (hSharedData);
	return "0";
}

string SystemInfo::getMBMLoaded(bool *dynamic)
{
	HANDLE memHandle = OpenFileMapping(FILE_MAP_READ, FALSE, "$M$B$M$5$S$D$");
	string retval = "0";

	if(memHandle)
	{
		retval = "1";
		CloseHandle(memHandle);
	}

	if(dynamic) *dynamic = true;
	return retval;
}

string SystemInfo::getNetIn(const vector<string> &arguments, bool *dynamic)
{
	int units = unitsDefault;

	if(arguments.size() > 0)
		units = atoi(arguments[0].c_str());

	if(dynamic) *dynamic = true;
	return formatByteSize(netStats->GetBytesPerSecIn(), 0, units);
}

string SystemInfo::getNetOut(const vector<string> &arguments, bool *dynamic)
{
	int units = unitsDefault;

	if(arguments.size() > 0)
		units = atoi(arguments[0].c_str());

	if(dynamic) *dynamic = true;
	return formatByteSize(netStats->GetBytesPerSecOut(), 0, units);
}

string SystemInfo::hideIfEmpty(const string &s, Label *label, bool *dynamic)
{
	if(label)
	{
		if(s.empty())
			label->internalHide();
		else if(!s.empty())
			label->internalShow();
	}
	
	return s;
}

string SystemInfo::empty(const string &s, bool *dynamic)
{
	return s.empty() ? "1" : "0";
}

string SystemInfo::notEmpty(const string &s, bool *dynamic)
{
	return s.empty() ? "0" : "1";
}

string SystemInfo::after(const string &s, const string &search)
{
	int i = s.find(search);
	i = (i < 0) ? 0 : i + search.length();
	return s.substr(i, s.length() - i);
}

string SystemInfo::afterLast(const string &s, const string &search)
{
	int i = s.rfind(search);
	i = (i < 0) ? 0 : i + search.length();
	return s.substr(i, s.length() - i);
}

string SystemInfo::before(const string &s, const string &search)
{
	int i = s.find(search);
	i = (i < 0) ? s.length() : i;
	return s.substr(0, i);
}

string SystemInfo::beforeLast(const string &s, const string &search)
{
	int i = s.rfind(search);
	i = (i < 0) ? s.length() : i;
	return s.substr(0, i);
}

string SystemInfo::between(const string &s, const string &left, const string &right)
{
	int start = s.find(left);
	start = (start < 0) ? 0 : start + left.length();

	int end = s.rfind(right);
	end = (end < 0) ? s.length() - 1 : end - right.length();

	return (start > end) ? "" : s.substr(start, end - start + 1);
}

string SystemInfo::capitalize(const string &s)
{
	string result;
	
	int length = s.length();
	int i = 0;
	
	while(i < length)
		result.append(1, (i == 0) ? toupper(s[i++]) : tolower(s[i++]));
	
	return result;
}

string SystemInfo::lowerCase(const string &s)
{
	string result;
	
	int length = s.length();
	int i = 0;
	
	while(i < length)
		result.append(1, tolower(s[i++]));
	
	return result;
}

string SystemInfo::trim(const string &s)
{
	int start = 0;
	int end = s.length() - 1;
	
	while(start <= end && isspace(s[start]))
		start++;
	
	while(start <= end && isspace(s[end]))
		end--;
	
	return s.substr(start, end - start + 1);
}

string SystemInfo::upperCase(const string &s)
{
	string result;
	
	int length = s.length();
	int i = 0;
	
	while(i < length)
		result.append(1, toupper(s[i++]));
	
	return result;
}

string SystemInfo::verticaldown(const string &s)
{
	string result;
	
	int length = s.length();
	int i = 0;
	
	while(i < length)
	{
		char ch = '\n';
		result.append(1, s[i]);
		result.append(1, ch);
		i++;
	}
	
	return result;
}

string SystemInfo::verticalup(const string &s)
{
	string result;
	
	int i = s.length()-1;
	
	while(i >= 0)
	{
		char ch = '\n';
		result.append(1, ch);
		result.append(1, s[i]);
		i--;
	}
	
	return result;
}

void SystemInfo::getDiskFreeSpace(const string &drive, largeInt &freeBytes, largeInt &totalBytes)
{
	static BOOL (WINAPI *pfnGetDiskFreeSpaceEx)(LPCSTR, PULARGE_INTEGER, PULARGE_INTEGER, PULARGE_INTEGER);
	static BOOL fChecked = FALSE;

	if(!fChecked)
	{
		pfnGetDiskFreeSpaceEx = (BOOL (WINAPI *)(LPCSTR, PULARGE_INTEGER, PULARGE_INTEGER, PULARGE_INTEGER)) GetProcAddress(
			GetModuleHandle("KERNEL32.DLL"), "GetDiskFreeSpaceExA");

		fChecked = TRUE;
	}

	if(pfnGetDiskFreeSpaceEx)
	{
		largeInt dummy;

		pfnGetDiskFreeSpaceEx(drive.c_str(),
			(PULARGE_INTEGER) &freeBytes,
			(PULARGE_INTEGER) &totalBytes,
			(PULARGE_INTEGER) &dummy);
	}
	else
	{
		DWORD sectorsPerCluster;
		DWORD bytesPerSector;
		DWORD freeClusters;
		DWORD totalClusters;

		GetDiskFreeSpace(drive.c_str(),
			&sectorsPerCluster,
			&bytesPerSector,
			&freeClusters,
			&totalClusters);

		largeInt bytesPerCluster = (largeInt) bytesPerSector * (largeInt) sectorsPerCluster;

		freeBytes = (largeInt) freeClusters * bytesPerCluster;
		totalBytes = (largeInt) totalClusters * bytesPerCluster;
	}
}

static const char *units[] = {
	"bytes", "KB", "MB", "GB", "TB", "PB", "EB"
};

string SystemInfo::formatByteSize(largeInt byteSize, largeInt total, int unit)
{
	char buffer[32];

	if(unit == unitsPercent) {
		if(total == 0) {
			return "0%";
		}
		else {
			StringCchPrintf(buffer, 32, "%d", (byteSize * 100) / total);
			return buffer;
		}
	}

	/* if(unit == unitsKB) {
		sprintf(buffer, "%lf", (double) (__int64) byteSize / 1024.0);
		return string(buffer);
	}
	else if(unit == unitsMB) {
		sprintf(buffer, "%lf", (double) (__int64)  byteSize / (1024.0 * 1024.0));
		return string(buffer);
	}
	else if(unit == unitsGB) {
		sprintf(buffer, "%lf", (double) (__int64)  byteSize / (1024.0 * 1024.0 * 1024.0));
		return string(buffer);
	}
	else if(unit == unitsPercent) {
		sprintf(buffer, "%d", (byteSize * 100) / total);
		return string(buffer);
	} */

	largeInt i = 0;
	bool addUnits = true;

	if(unit == unitsBytes) {
		i = 0;
		addUnits = false;
	}
	if(unit == unitsKB) {
		i = 1;
		addUnits = false;
	}
	else if(unit == unitsMB) {
		i = 2;
		addUnits = false;
	}
	else if(unit == unitsGB) {
		i = 3;
		addUnits = false;
	}
	else {
		if(byteSize & 0xF000000000000000) i++;
		if(byteSize & 0xFFFC000000000000) i++;
		if(byteSize & 0xFFFFFF0000000000) i++;
		if(byteSize & 0xFFFFFFFFC0000000) i++;
		if(byteSize & 0xFFFFFFFFFFF00000) i++;
		if(byteSize & 0xFFFFFFFFFFFFFC00) i++;
	}

	largeInt divisor = 1 << (i * 10);
	largeInt quotient = byteSize / divisor;
	largeInt remainder = ((byteSize % divisor) * 100) / divisor;

	if(quotient > 9 && quotient < 100)
		remainder = remainder / 10;
	else if(quotient >= 100)
		remainder = 0;

	if(remainder > 0)
		StringCchPrintf(buffer, 32, "%d.%02d", (int) quotient, (int) remainder);
	else
		StringCchPrintf(buffer, 32, "%d", (int) quotient);

	if(addUnits) {
		StringCchCat(buffer, 32, " ");
		StringCchCat(buffer, 32, units[i]);
	}

	return string(buffer);
}

static inline int isSequenceChar(int ch)
{
	return (ch == 'd' || ch == 'm' || ch == 'y' || ch == 'h' || ch == 'i' || ch == 'n' || ch == 's');
}

string SystemInfo::formatDateTime(const string &format, const SYSTEMTIME &st, int span)
{
	const char *formatPtr = format.c_str();
	string output;
	int lastChar = 0;
	int sequenceChar = 0;
	int count = 0;
	int ch = *formatPtr++;
	
	while(ch || sequenceChar)
	{
		if(sequenceChar)
		{
			if(ch == sequenceChar)
			{
				count++;
			}
			else if(isalpha(ch))
			{
				output.append(count, sequenceChar);
				output.append(1, ch);
				
				sequenceChar = 0;
				count = 0;
			}
			else
			{
				switch(sequenceChar)
				{
					case 'd':
					{
						switch(count)
						{
							case 1:
							{
								char temp[8];
								StringCchPrintf(temp, 8, "%d", st.wDay);
								output.append(temp);
								break;
							}
							
							case 2:
							{
								char temp[8];
								StringCchPrintf(temp, 8, "%02d", st.wDay);
								output.append(temp);
								break;
							}
							
							case 3:
							{
								if(st.wDayOfWeek < 7)
									output.append(shortWeekdays[st.wDayOfWeek]);
								
								break;
							}
							
							case 4:
							{
								if(st.wDayOfWeek < 7)
									output.append(weekdays[st.wDayOfWeek]);
									
								break;
							}
							
							default:
							{
								output.append(count, sequenceChar);
								break;
							}
						}
						
						break;
					}
					
					case 'm':
					{
						if(span) break;

						switch(count)
						{
							case 1:
							{
								char temp[8];
								StringCchPrintf(temp, 8, "%d", st.wMonth);
								output.append(temp);
								break;
							}
							
							case 2:
							{
								char temp[8];
								StringCchPrintf(temp, 8, "%02d", st.wMonth);
								output.append(temp);
								break;
							}
							
							case 3:
							{
								if(st.wMonth > 0 && st.wMonth < 13)
									output.append(shortMonths[st.wMonth - 1]);
								
								break;
							}
							
							case 4:
							{
								if(st.wMonth > 0 && st.wMonth < 13)
									output.append(months[st.wMonth - 1]);
								
								break;
							}
							
							default:
							{
								output.append(count, sequenceChar);
								break;
							}
						}
						
						break;
					}
					
					case 'y':
					{
						if(span) break;

						if(count == 2 || count == 4)
						{
							char temp[8];
							StringCchPrintf(temp, 8, "%04d", st.wYear);
							output.append((count == 2) ? temp + 2 : temp);
						}
						else
						{
							output.append(count, sequenceChar);
						}
						
						break;
					}
					
					case 'h':
					{
						if(count == 1 || count == 2)
						{
							char temp[8];
							int hour = span ? st.wHour : (st.wHour == 0) ? 12 : (st.wHour > 12) ? st.wHour - 12 : st.wHour;
							StringCchPrintf(temp, 8, (count == 1) ? "%d" : "%02d", hour);
							output.append(temp);
						}
						else
						{
							output.append(count, sequenceChar);
						}
						
						break;
					}
					
					case 'i':
					{
						if(count == 1 || count == 2)
						{
							char temp[8];
							StringCchPrintf(temp, 8, (count == 1) ? "%d" : "%02d", st.wHour);
							output.append(temp);
						}
						else
						{
							output.append(count, sequenceChar);
						}
						
						break;
					}
					
					case 'n':
					{
						if(count == 1 || count == 2)
						{
							char temp[8];
							StringCchPrintf(temp, 8, (count == 1) ? "%d" : "%02d", st.wMinute);
							output.append(temp);
						}
						else
						{
							output.append(count, sequenceChar);
						}
						
						break;
					}
					
					case 's':
					{
						if(count == 1 || count == 2)
						{
							char temp[8];
							StringCchPrintf(temp, 8, (count == 1) ? "%d" : "%02d", st.wSecond);
							output.append(temp);
						}
						else
						{
							output.append(count, sequenceChar);
						}
						
						break;
					}
					
					default:
					{
						output.append(count, sequenceChar);
						break;
					}
				}
				
				if(ch) output.append(1, ch);
				sequenceChar = 0;
				count = 0;
			}
		}
		else
		{
			if(isSequenceChar(ch) && !isalpha(lastChar))
			{
				sequenceChar = ch;
				count = 1;
			}
			else if(ch == 'a' && !isalpha(lastChar))
			{
				const char *p = formatPtr - 1;
				
				// special case: am/pm
				if(_strnicmp(p, "am/pm", 5) == 0)
				{
					output.append((st.wHour >= 12) ? "pm" : "am");
					formatPtr = formatPtr + 4;
				}
				else
				{
					output.append(1, ch);
				}
			}
			else
			{
				output.append(1, ch);
			}
		}
		
		lastChar = ch;
		if(ch) ch = *formatPtr++;
	}
	
	return output;
}

void SystemInfo::getLocalizedTime(SYSTEMTIME *pst, const string &timezone)
{
	string temp = lowerCase(timezone);
	const char *s = temp.c_str();

	if(s[0] == 'u' && s[1] == 't' && s[2] == 'c')
	{
		s = s + 3;
		GetSystemTime(pst);
	}
	else
	{
		GetLocalTime(pst);
	}

	while(s[0] == ' ')
		s++;

	int dHour = 0;
	int dMinute = 0;
	int sign = 1;

	if(s[0] == '+')
	{
		sign = 1;
		s++;
	}
	else if(s[0] == '-')
	{
		sign = -1;
		s++;
	}

	while(s[0] >= '0' && s[0] <= '9')
	{
		dHour = (dHour * 10) + (s[0] - '0');
		s++;
	}

	if(s[0] == ':')
	{
		s++;

		while(s[0] >= '0' && s[0] <= '9')
		{
			dMinute = (dMinute * 10) + (s[0] - '0');
			s++;
		}
	}

	largeInt base;
	largeInt delta = (largeInt) (sign * ((dHour * 60) + dMinute)) * 600000000;

	SystemTimeToFileTime(pst, (LPFILETIME) &base);
	base = base + delta;
	FileTimeToSystemTime((LPFILETIME) &base, pst);
}

char *SystemInfo::getLocaleInfo(LCTYPE type)
{
	int len = GetLocaleInfo(LOCALE_USER_DEFAULT, type, 0, 0);
	char *buffer = new char[len];
	GetLocaleInfo(LOCALE_USER_DEFAULT, type, buffer, len);
	return buffer;
}

string SystemInfo::getClipboardText(bool *dynamic)
{
	char buffer[256];
	string result = "";

	GetClipboardText(NULL, buffer, 256);
	result = buffer;

	if(dynamic)
		*dynamic = true;

	return result;
}

SystemInfo *systemInfo;
